==> v1.43.0.zip: google.golang.org/genai@v1.43.0/tokens.go <==
// Copyright 2025 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package genai

import (
	"context"
	"fmt"
	"log"
	"net/http"
	"reflect"
	"strings"
	"sync"
)

// getFieldMasks returns a comma-separated list of fields to be used in the field mask.
func getFieldMasks(setup map[string]any) string {
	var fields []string

	for k, v := range setup {
		// Check if v is a non-empty map (2nd layer recursion)
		if vMap, ok := v.(map[string]any); ok && len(vMap) > 0 {
			// Note: The Python snippet had `field = [f'.' for kk in v.keys()]`.
			// This implies a string format. Standard FieldMask is "parent.child".
			// We assume the intent was f"{}.{}".
			for kk := range vMap {
				fields = append(fields, fmt.Sprintf("%s.%s", k, kk))
			}
		} else {
			// 1st layer
			fields = append(fields, k)
		}
	}

	return strings.Join(fields, ",")
}

// isGenerationConfigField checks if the field is a generationConfig field.
func isGenerationConfigField(field string, generationConfigList []string) bool {

	for _, generationConfigField := range generationConfigList {
		if field == generationConfigField {
			return true
		}
		// Special case for passing replay tests. The table test case recorded top_k in a long string
		// for field mask that can't be easily convert to camel case. The actual Go CUJs, it will be
		// all in camel cases.
		if field == "top_k" {
			return true
		}
	}
	return false
}

func fieldMaskList(fieldMaskRaw any, exists bool) []string {
	var fieldMaskSlice []string
	if exists {
		switch v := fieldMaskRaw.(type) {
		case []string:
			fieldMaskSlice = v
		case []interface{}:
			for _, item := range v {
				if s, ok := item.(string); ok {
					fieldMaskSlice = append(fieldMaskSlice, s)
				}
			}
		}
	}
	return fieldMaskSlice
}

// Convert BidiGenerateContentSetup to token setup.
func ConvertBidiSetupToTokenSetup(
	body map[string]any,
	config *CreateAuthTokenConfig,
) map[string]any {

	// Extract bidiGenerateContentSetup
	bidiSetup, hasBidi := body["bidiGenerateContentSetup"].(map[string]any)

	// Check if bidiSetup exists and has 'setup'
	var setup map[string]any
	if hasBidi {
		if s, ok := bidiSetup["setup"].(map[string]any); ok {
			setup = s
		}
	}

	if setup != nil {
		// Handling mismatch between AuthToken service and BidiGenerateContent service
		body["bidiGenerateContentSetup"] = setup

		// Convert non null bidiGenerateContentSetup to field_mask
		fieldMask := getFieldMasks(setup)

		// Check config.LockAdditionalFields state

		if config == nil || config.LockAdditionalFields == nil {
			// Case 1: LockAdditionalFields is None (nil). Global lock. Unset fieldMask
			delete(body, "fieldMask")

		} else if config != nil && len(config.LockAdditionalFields) == 0 {
			// Case 2: LockAdditionalFields is explicit False
			body["fieldMask"] = fieldMask

		} else {
			// Case 3: LockAdditionalFields is True (implied by else).
			// Lock non null + additional fields

			// Check if there is an existing fieldMask to merge
			fieldMaskRaw, exists := body["fieldMask"]
			additionalFieldsList := fieldMaskList(fieldMaskRaw, exists)

			var fieldMaskList []string

			if len(additionalFieldsList) > 0 {
				// Get all fields in GenerationConfig
				var generationConfigFields []string
				t := reflect.TypeOf(GenerationConfig{})

				for i := 0; i < t.NumField(); i++ {
					field := t.Field(i)
					tag := field.Tag.Get("json")
					key, _, _ := strings.Cut(tag, ",")

					if key != "" && key != "-" {
						generationConfigFields = append(generationConfigFields, key)
					}
				}
				for _, field := range additionalFieldsList {
					// Logic: if field in generation_config_list -> prefix with "generationConfig."
					processedField := field
					if isGenerationConfigField(field, generationConfigFields) {
						processedField = fmt.Sprintf("generationConfig.%s", field)
					}
					fieldMaskList = append(fieldMaskList, processedField)
				}
				body["fieldMask"] = fieldMask + "," + strings.Join(fieldMaskList, ",")
			} else {
				// Lock all fields
				delete(body, "fieldMask")
			}
		}
	} else {
		// No bidi setup found or no 'setup' key inside it
		fieldMaskRaw, exists := body["fieldMask"]

		fieldMaskSlice := fieldMaskList(fieldMaskRaw, exists)

		fieldMaskStr := strings.Join(fieldMaskSlice, ",")
		if len(fieldMaskSlice) > 0 {
			body["fieldMask"] = fieldMaskStr
		} else {
			delete(body, "fieldMask")
		}
	}

	// Final cleanup: Check if bidiGenerateContentSetup exists/is truthy
	// In Go maps, we check existence. Python checks truthiness (dict is not empty).
	if val, ok := body["bidiGenerateContentSetup"]; !ok || val == nil {
		delete(body, "bidiGenerateContentSetup")
	} else if vMap, ok := val.(map[string]any); ok && len(vMap) == 0 {
		// Handle empty dict case equivalent to Python `if not ...`
		delete(body, "bidiGenerateContentSetup")
	}

	return body
}

var experimentalWarningTokensCreate sync.Once

// Tokens provides methods for managing the context caching.
// You don't need to initiate this struct. Create a client instance via NewClient, and
// then access Tokens through client.Tokens field.
type Tokens struct {
	apiClient *apiClient
}

// Create creates a new cached content resource.
func (m Tokens) Create(ctx context.Context, config *CreateAuthTokenConfig) (*AuthToken, error) {
	experimentalWarningTokensCreate.Do(func() {
		log.Println("The SDK's ephemeral tokens implementation is experimental, and may change in future versions.")
	})

	parameterMap := make(map[string]any)

	kwargs := map[string]any{"config": config}
	err := deepMarshal(kwargs, &parameterMap)
	if err != nil {
		return nil, err
	}

	var httpOptions *HTTPOptions
	if config == nil || config.HTTPOptions == nil {
		httpOptions = &HTTPOptions{}
	} else {
		httpOptions = config.HTTPOptions
	}
	if httpOptions.Headers == nil {
		httpOptions.Headers = http.Header{}
	}
	var response = new(AuthToken)
	var responseMap map[string]any
	var toConverter func(*apiClient, map[string]any, map[string]any) (map[string]any, error)
	if m.apiClient.clientConfig.Backend == BackendVertexAI {
		return nil, fmt.Errorf("method Create is only supported in the Gemini Developer client. You can choose to use Gemini Developer client by setting ClientConfig.Backend to BackendGeminiAPI.")

	} else {
		toConverter = createAuthTokenParametersToMldev

	}

	body, err := toConverter(m.apiClient, parameterMap, nil)
	if err != nil {
		return nil, err
	}
	var path string
	var urlParams map[string]any
	if _, ok := body["_url"]; ok {
		urlParams = body["_url"].(map[string]any)
		delete(body, "_url")
	}
	path, err = formatMap("auth_tokens", urlParams)

	if err != nil {
		return nil, fmt.Errorf("invalid url params: %#v.\n%w", urlParams, err)
	}
	if _, ok := body["_query"]; ok {
		query, err := createURLQuery(body["_query"].(map[string]any))
		if err != nil {
			return nil, err
		}
		path += "?" + query
		delete(body, "_query")
	}
	transformedBody := ConvertBidiSetupToTokenSetup(body, config)
	delete(transformedBody, "config")

	responseMap, err = sendRequest(ctx, m.apiClient, path, http.MethodPost, transformedBody, httpOptions)
	if err != nil {
		return nil, err
	}
	err = mapToStruct(responseMap, response)
	if err != nil {
		return nil, err
	}

	if field, ok := reflect.TypeOf(response).Elem().FieldByName("SDKHTTPResponse"); ok {
		{
			if reflect.ValueOf(response).Elem().FieldByName("SDKHTTPResponse").IsValid() {
				{
					reflect.ValueOf(response).Elem().FieldByName("SDKHTTPResponse").Set(reflect.Zero(field.Type))
				}
			}
		}
	}

	return response, nil
}

==> v1.45.0.zip: google.golang.org/genai@v1.45.0/tokens.go <==
// Copyright 2025 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package genai

import (
	"context"
	"fmt"
	"log"
	"net/http"
	"reflect"
	"strings"
	"sync"
)

// getFieldMasks returns a comma-separated list of fields to be used in the field mask.
func getFieldMasks(setup map[string]any) string {
	var fields []string

	for k, v := range setup {
		// Check if v is a non-empty map (2nd layer recursion)
		if vMap, ok := v.(map[string]any); ok && len(vMap) > 0 {
			// Note: The Python snippet had `field = [f'.' for kk in v.keys()]`.
			// This implies a string format. Standard FieldMask is "parent.child".
			// We assume the intent was f"{}.{}".
			for kk := range vMap {
				fields = append(fields, fmt.Sprintf("%s.%s", k, kk))
			}
		} else {
			// 1st layer
			fields = append(fields, k)
		}
	}

	return strings.Join(fields, ",")
}

// isGenerationConfigField checks if the field is a generationConfig field.
func isGenerationConfigField(field string, generationConfigList []string) bool {

	for _, generationConfigField := range generationConfigList {
		if field == generationConfigField {
			return true
		}
		// Special case for passing replay tests. The table test case recorded top_k in a long string
		// for field mask that can't be easily convert to camel case. The actual Go CUJs, it will be
		// all in camel cases.
		if field == "top_k" {
			return true
		}
	}
	return false
}

func fieldMaskList(fieldMaskRaw any, exists bool) []string {
	var fieldMaskSlice []string
	if exists {
		switch v := fieldMaskRaw.(type) {
		case []string:
			fieldMaskSlice = v
		case []interface{}:
			for _, item := range v {
				if s, ok := item.(string); ok {
					fieldMaskSlice = append(fieldMaskSlice, s)
				}
			}
		}
	}
	return fieldMaskSlice
}

// Convert BidiGenerateContentSetup to token setup.
func ConvertBidiSetupToTokenSetup(
	body map[string]any,
	config *CreateAuthTokenConfig,
) map[string]any {

	// Extract bidiGenerateContentSetup
	bidiSetup, hasBidi := body["bidiGenerateContentSetup"].(map[string]any)

	// Check if bidiSetup exists and has 'setup'
	var setup map[string]any
	if hasBidi {
		if s, ok := bidiSetup["setup"].(map[string]any); ok {
			setup = s
		}
	}

	if setup != nil {
		// Handling mismatch between AuthToken service and BidiGenerateContent service
		body["bidiGenerateContentSetup"] = setup

		// Convert non null bidiGenerateContentSetup to field_mask
		fieldMask := getFieldMasks(setup)

		// Check config.LockAdditionalFields state

		if config == nil || config.LockAdditionalFields == nil {
			// Case 1: LockAdditionalFields is None (nil). Global lock. Unset fieldMask
			delete(body, "fieldMask")

		} else if config != nil && len(config.LockAdditionalFields) == 0 {
			// Case 2: LockAdditionalFields is explicit False
			body["fieldMask"] = fieldMask

		} else {
			// Case 3: LockAdditionalFields is True (implied by else).
			// Lock non null + additional fields

			// Check if there is an existing fieldMask to merge
			fieldMaskRaw, exists := body["fieldMask"]
			additionalFieldsList := fieldMaskList(fieldMaskRaw, exists)

			var fieldMaskList []string

			if len(additionalFieldsList) > 0 {
				// Get all fields in GenerationConfig
				var generationConfigFields []string
				t := reflect.TypeOf(GenerationConfig{})

				for i := 0; i < t.NumField(); i++ {
					field := t.Field(i)
					tag := field.Tag.Get("json")
					key, _, _ := strings.Cut(tag, ",")

					if key != "" && key != "-" {
						generationConfigFields = append(generationConfigFields, key)
					}
				}
				for _, field := range additionalFieldsList {
					// Logic: if field in generation_config_list -> prefix with "generationConfig."
					processedField := field
					if isGenerationConfigField(field, generationConfigFields) {
						processedField = fmt.Sprintf("generationConfig.%s", field)
					}
					fieldMaskList = append(fieldMaskList, processedField)
				}
				body["fieldMask"] = fieldMask + "," + strings.Join(fieldMaskList, ",")
			} else {
				// Lock all fields
				delete(body, "fieldMask")
			}
		}
	} else {
		// No bidi setup found or no 'setup' key inside it
		fieldMaskRaw, exists := body["fieldMask"]

		fieldMaskSlice := fieldMaskList(fieldMaskRaw, exists)

		fieldMaskStr := strings.Join(fieldMaskSlice, ",")
		if len(fieldMaskSlice) > 0 {
			body["fieldMask"] = fieldMaskStr
		} else {
			delete(body, "fieldMask")
		}
	}

	// Final cleanup: Check if bidiGenerateContentSetup exists/is truthy
	// In Go maps, we check existence. Python checks truthiness (dict is not empty).
	if val, ok := body["bidiGenerateContentSetup"]; !ok || val == nil {
		delete(body, "bidiGenerateContentSetup")
	} else if vMap, ok := val.(map[string]any); ok && len(vMap) == 0 {
		// Handle empty dict case equivalent to Python `if not ...`
		delete(body, "bidiGenerateContentSetup")
	}

	return body
}

var experimentalWarningTokensCreate sync.Once

// Tokens provides methods for managing the context caching.
// You don't need to initiate this struct. Create a client instance via NewClient, and
// then access Tokens through client.Tokens field.
type Tokens struct {
	apiClient *apiClient
}

// Create creates a new cached content resource.
func (m Tokens) Create(ctx context.Context, config *CreateAuthTokenConfig) (*AuthToken, error) {
	experimentalWarningTokensCreate.Do(func() {
		log.Println("The SDK's ephemeral tokens implementation is experimental, and may change in future versions.")
	})

	parameterMap := make(map[string]any)

	kwargs := map[string]any{"config": config}
	err := deepMarshal(kwargs, &parameterMap)
	if err != nil {
		return nil, err
	}

	var httpOptions *HTTPOptions
	if config == nil || config.HTTPOptions == nil {
		httpOptions = &HTTPOptions{}
	} else {
		httpOptions = config.HTTPOptions
	}
	if httpOptions.Headers == nil {
		httpOptions.Headers = http.Header{}
	}
	var response = new(AuthToken)
	var responseMap map[string]any
	var toConverter func(*apiClient, map[string]any, map[string]any, map[string]any) (map[string]any, error)
	if m.apiClient.clientConfig.Backend == BackendVertexAI {
		return nil, fmt.Errorf("method Create is only supported in the Gemini Developer client. You can choose to use Gemini Developer client by setting ClientConfig.Backend to BackendGeminiAPI.")

	} else {
		toConverter = createAuthTokenParametersToMldev

	}

	body, err := toConverter(m.apiClient, parameterMap, nil, parameterMap)
	if err != nil {
		return nil, err
	}
	var path string
	var urlParams map[string]any
	if _, ok := body["_url"]; ok {
		urlParams = body["_url"].(map[string]any)
		delete(body, "_url")
	}
	path, err = formatMap("auth_tokens", urlParams)

	if err != nil {
		return nil, fmt.Errorf("invalid url params: %#v.\n%w", urlParams, err)
	}
	if _, ok := body["_query"]; ok {
		query, err := createURLQuery(body["_query"].(map[string]any))
		if err != nil {
			return nil, err
		}
		path += "?" + query
		delete(body, "_query")
	}
	transformedBody := ConvertBidiSetupToTokenSetup(body, config)
	delete(transformedBody, "config")

	responseMap, err = sendRequest(ctx, m.apiClient, path, http.MethodPost, transformedBody, httpOptions)
	if err != nil {
		return nil, err
	}
	err = mapToStruct(responseMap, response)
	if err != nil {
		return nil, err
	}

	if field, ok := reflect.TypeOf(response).Elem().FieldByName("SDKHTTPResponse"); ok {
		{
			if reflect.ValueOf(response).Elem().FieldByName("SDKHTTPResponse").IsValid() {
				{
					reflect.ValueOf(response).Elem().FieldByName("SDKHTTPResponse").Set(reflect.Zero(field.Type))
				}
			}
		}
	}

	return response, nil
}
